#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include "stringlib.h"
#include "utils.h"



typedef struct string{
    char* str;
    unsigned long long size;
    unsigned long long capacity;
} string_t;


string_t* string_new(){
    char* empty_string = (char*) malloc(MIN_STRING_CAPACITY*sizeof(char));
    string_t* new_str = malloc(sizeof(string_t));
    new_str->str = empty_string;
    new_str->size = 0;
    new_str->capacity = MIN_STRING_CAPACITY;
    return new_str;
}

string_t* string_from_chars(const char* str){
    const unsigned long long length = strlen(str);
    const unsigned long long capacity = umax(next_power_of_2(length), MIN_STRING_CAPACITY);
    char* new_chars = malloc(capacity*sizeof(char));
    strncpy(new_chars, str, capacity);
    string_t* new_str = malloc(sizeof(string_t));
    new_str->str = new_chars;
    new_str->size = length;
    new_str->capacity = capacity;
    return new_str;
}

char* string_cpy_to_chars(const string_t* str){
    char* new_chars = malloc(((str->size)+1)*sizeof(char));
    for(unsigned long long i = 0; i < str->size; ++i){
        new_chars[i] = str->str[i];
    }
    new_chars[str->size] = '\0';

    return new_chars;
}

string_t* string_copy(const string_t* str){
    string_t* new_str = string_new();
    char* chars = malloc((str->capacity)*sizeof(char));
    for(unsigned long long i = 0; i < str->size; ++i){
        chars[i] = str->str[i];
    }
    new_str->str = chars;
    new_str->capacity = str->capacity;
    new_str->size = str->size;
    return new_str;
}

const char* string_cstr(string_t* str){
    string_set_min_capacity(str, str->size+1);
    str->str[str->size] = '\0';
    return str->str;
}

unsigned long long string_get_size(const string_t* str){
    return str->size;
}

unsigned long long string_get_capacity(const string_t* str){
    return str->capacity;
}

void string_set_min_capacity(string_t* str, const unsigned long long size){
    if(size > str->capacity){
        char* new_str = realloc(str->str, size*sizeof(char));
        if(new_str==NULL){
            exit(2);
        }
        str->str = new_str;
        str->capacity = size;    
    }
    return;
}

void string_append(string_t* dest, const string_t* source){
    const unsigned long long dest_size = dest->size;
    const unsigned long long source_size = source->size;
    string_set_min_capacity(dest, umax(next_power_of_2(dest_size+source_size), MIN_STRING_CAPACITY));
    for(unsigned long long i = 0; i < source_size; ++i){
        dest->str[dest_size+i] = source->str[i];
    }
    dest->size = dest_size+source_size;
    return;
}

void string_append_char(string_t* dest, const char c){
    const unsigned long long size = dest->size;
    string_set_min_capacity(dest, umax(next_power_of_2(size+1), MIN_STRING_CAPACITY));
    dest->str[size] = c;
    dest->size = size+1;
    return;
}

void string_append_chars(string_t* dest, const char* source){
    string_t* temp = string_from_chars(source);
    string_append(dest, temp);
    string_free(temp);
    return;
}

void string_print(const string_t* str){
    for(unsigned long long i = 0; i < str->size; ++i){
        printf("%c", str->str[i]);
    }
    return;
}

string_t* string_substr(const string_t* str, const unsigned long long position, const unsigned long long length){
    string_t* new_str = string_new();
    if(position >= str->size){
        return new_str;
    }
    unsigned long long actual_size = min(length, str->size - position);
    string_set_min_capacity(new_str, umax(next_power_of_2(actual_size), MIN_STRING_CAPACITY));
    for(unsigned long long i = 0; i < actual_size; ++i){
        new_str->str[i] = str->str[position+i];
    }
    new_str->size = actual_size;
    return new_str;
}

void string_erase(string_t* str, const unsigned long long position, const unsigned long long length){
    if(position >= str->size){
        return;
    }
    unsigned long long actual_size = min(length, str->size - position);
    char* cpy = string_cpy_to_chars(str);

    for(unsigned long long i = 0; i < position; ++i){
        str->str[i] = cpy[i];
    }
    for(unsigned long long i = position + actual_size; i < str->size; ++i){
        str->str[i-actual_size] = cpy[i];
    }
    str->size -= actual_size;
    free(cpy);
    return;
}

void string_clear(string_t* str){
    str->size = 0;
}

void string_free(string_t* str){
    //According to valgrind strings made from string_copy aren't free'd correctly, but everything I tested seemed like it free'd correctly. Must be a valgrind bug...
    free(str->str);
    free(str);
    return;
}